# This file is part of tridiax, a toolkit for solving tridiagonal systems. tridiax is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax.numpy as jnp
from jax import lax
from jax.numpy import ndarray


def thomas_solve(
    lower: ndarray,
    diag: ndarray,
    upper: ndarray,
    solve: ndarray,
    divide_last: bool = True,
) -> ndarray:
    """
    Return solution `x` to tridiagonal system `Ax = b` with Thomas algorithm.

    Args:
        divide_last: If `True`, the last row will be divided through by the diagonal
            such that the entire diagonal is 1. If `False`, this step will be skipped
            only for the very last value on the diagonal.
    """
    u, u_upper, y = thomas_triang_upper(
        lower, diag, upper, solve, divide_last=divide_last
    )
    return thomas_backsub_lower(y, u_upper, u)


def thomas_triang_lower(
    lower: jnp.ndarray,
    diag: jnp.ndarray,
    upper: jnp.ndarray,
    solve: jnp.ndarray,
    divide_last: bool = False,
):
    """
    Note that `divide_last=False` by default here, unlike in `thomas_solve`. This is to
    support this using `thomas_triang` in `Jaxley`.
    """
    n = len(solve)
    # Triangulation is only needed for matrizes, not scalars.
    if n > 1:
        u_upper = jnp.zeros(n - 1)
        u_upper = u_upper.at[0].set(upper[0] / diag[0])
        u_upper = lax.fori_loop(
            1, n - 1, _u_upper_update, (u_upper, upper, diag, lower)
        )[0]

        y = jnp.zeros(n)
        y = y.at[0].set(solve[0] / diag[0])
        y = lax.fori_loop(1, n - 1, _y_update, (solve, lower, y, diag, u_upper))[2]

        # Handle last row separately.
        newdiag = jnp.ones_like(diag)
        y = y.at[-1].set(solve[-1] - lower[-1] * y[-2])
        if divide_last:
            y = y.at[-1].set(y[-1] / (diag[-1] - lower[-1] * u_upper[-1]))
        else:
            newdiag = newdiag.at[-1].set(diag[-1] - lower[-1] * u_upper[-1])

        return newdiag, u_upper, y
    else:
        return diag, upper, solve


def thomas_triang_upper(
    lower: jnp.ndarray,
    diag: jnp.ndarray,
    upper: jnp.ndarray,
    solve: jnp.ndarray,
    divide_last: bool = False,
):
    """Triangulate system by removing the upper diagonal."""
    diag, lower, y = thomas_triang_lower(
        lower=jnp.flip(upper),
        diag=jnp.flip(diag),
        upper=jnp.flip(lower),
        solve=jnp.flip(solve),
        divide_last=divide_last,
    )
    return jnp.flip(diag), jnp.flip(lower), jnp.flip(y)


def thomas_backsub_upper(solve: jnp.ndarray, upper: jnp.ndarray, diag: jnp.ndarray):
    """Backsubstitute to remove the upper diagonal."""
    n = len(solve)
    x = jnp.zeros(n)
    x = x.at[0].set(solve[n - 1] / diag[n - 1])
    # This if-case should not be necessary, but even if fori does not do any
    # iterations, it still traces the body_fun and breaks without this if-case.
    if n > 1:
        x = lax.fori_loop(0, n - 1, _x_update, (solve, upper, x, n))[2]

    return jnp.flip(x)


def thomas_backsub_lower(solve: jnp.ndarray, lower: jnp.ndarray, diag: jnp.ndarray):
    """Backsubstitute to remove the lower diagonal."""
    solution = thomas_backsub_upper(jnp.flip(solve), jnp.flip(lower), jnp.flip(diag))
    return jnp.flip(solution)


def _u_upper_update(i, val):
    u_upper, upper, diag, lower = val
    u_upper = u_upper.at[i].set(upper[i] / (diag[i] - lower[i - 1] * u_upper[i - 1]))
    return (u_upper, upper, diag, lower)


def _y_update(i, val):
    solve, lower, y, diag, u_upper = val
    y = y.at[i].set(
        (solve[i] - lower[i - 1] * y[i - 1]) / (diag[i] - lower[i - 1] * u_upper[i - 1])
    )
    return (solve, lower, y, diag, u_upper)


def _x_update(i, val):
    y, u_upper, x, n = val
    x = x.at[i + 1].set(y[n - i - 2] - u_upper[n - i - 2] * x[i])
    return (y, u_upper, x, n)



# This file is part of tridiax, a toolkit for solving tridiagonal systems. tridiax is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from math import log2
from typing import Optional

import jax.numpy as jnp
from jax import vmap
from jax.numpy import ndarray


def divide_conquer_solve(
    lower: ndarray,
    diag: ndarray,
    upper: ndarray,
    solve: ndarray,
    indexing: Optional[ndarray] = None,
) -> ndarray:
    """
    Runs the divide and conquer algorithm to solve the tridiagonal system.

    Args:
        indexing: After having solved the linear system, the solution has to be
            rearraged into the correct order. If `indexing=None`, the reordering is
            done automatically, but this can take a few seconds. Instead, if
            `indexing` is passed, it is direclty used to reorder the result. This can
            be helpful if `divide_and_conquer` is applied successively for many systems
            of the same size. Use ```indexing = divide_conquer_index(dim)``` to
            precompute the indizes.
    """
    dim = len(diag)

    lower = jnp.expand_dims(lower, axis=0)
    diag = jnp.expand_dims(diag, axis=0)
    upper = jnp.expand_dims(upper, axis=0)
    solve = jnp.expand_dims(solve, axis=0)

    # Reduce the system.
    for _ in range(int(log2(dim))):
        a_bars, b_bars, c_bars, solve_bars = vmap(_reduce_system)(
            (lower, diag, upper, solve)
        )
        upper = _split(c_bars)
        diag = _split(b_bars)
        lower = _split(a_bars)
        solve = _split(solve_bars)

    # Solve the (now diagonal) system.
    x = solve / diag

    if indexing is None:
        return jnp.squeeze(_reorder(x, dim))
    else:
        return jnp.squeeze(x)[indexing]


def divide_conquer_index(dim: int):
    return _reorder(jnp.expand_dims(jnp.arange(0, dim), 1), dim)[0]


def _reorder(x, dim):
    """
    Reorder the solution.
    """
    for _ in range(int(log2(dim))):
        x = _merge_dims(x)
        x = vmap(_rearrange_dims, 0)(x)
    return x


def _reduce_system(vecs):
    lower, diag, upper, solve = vecs
    a = upper
    b = diag
    c = lower
    r = solve

    alpha1 = -a / b[1:]
    beta1 = -c / b[:-1]

    a_bar = alpha1[:-1] * a[1:]
    c_bar = beta1[1:] * c[:-1]

    b_bar = b.at[:-1].set(b[:-1] + alpha1 * c)
    b_bar = b_bar.at[1:].set(b_bar[1:] + beta1 * a)

    r_bar = r.at[:-1].set(r[:-1] + alpha1 * r[1:])
    r_bar = r_bar.at[1:].set(r_bar[1:] + beta1 * r[:-1])
    return (c_bar, b_bar, a_bar, r_bar)


def _split(vector):
    r = jnp.asarray([vector[:, ::2], vector[:, 1::2]])
    r2 = jnp.transpose(r, (1, 0, 2))
    return jnp.concatenate(r2, axis=0)


def _merge_dims(x):
    return jnp.reshape(x, (x.shape[0] // 2, x.shape[1] * 2))


def _rearrange_dims(x):
    dim = len(x)

    x_even = x[: -dim // 2]
    x_odd = x[dim // 2 :]

    x = x.at[::2].set(x_even)
    x = x.at[1::2].set(x_odd)
    return x



# This file is part of tridiax, a toolkit for solving tridiagonal systems. tridiax is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

VERSION = (0, 2, 0)

__version__ = ".".join(map(str, VERSION))



# This file is part of tridiax, a toolkit for solving tridiagonal systems. tridiax is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from tridiax.__version__ import __version__  # noqa: F401
from tridiax.divide_and_conquer import divide_conquer_index, divide_conquer_solve
from tridiax.stone import stone_solve
from tridiax.thomas import thomas_solve



# This file is part of tridiax, a toolkit for solving tridiagonal systems. tridiax is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from math import log
from typing import Tuple

import jax.numpy as jnp
import numpy as np


def stone_solve(
    lower: jnp.ndarray,
    diag: jnp.ndarray,
    upper: jnp.ndarray,
    solve: jnp.ndarray,
    stabilize: bool = True,
    optimized_lu: bool = True,
) -> jnp.ndarray:
    """
    Return solution `x` to tridiagonal system `Ax = b` with Stone's algorithm.

    Proceeds in three steps (just like Thomas algorithm):
    1) LU decomposition: LUx = b
    2) Triangularization: Solve Ly = b
    3) Backsubstitution: Solve Ux = y
    """
    dim = len(diag)

    # The stone solver only supports powers of two. Thus, we pad here.
    power_of_two = np.log2(dim)
    if not power_of_two.is_integer():
        dim_gap = int(2 ** np.ceil(power_of_two) - dim)
        ones_pad = jnp.ones(dim_gap)
        zeros_pad = jnp.zeros(dim_gap)
        diag = jnp.concatenate([diag, ones_pad])
        solve = jnp.concatenate([solve, ones_pad])
        upper = jnp.concatenate([upper, zeros_pad])
        lower = jnp.concatenate([lower, zeros_pad])

    u, upper, y = stone_triang_upper(
        lower, diag, upper, solve, stabilize=stabilize, optimized_lu=optimized_lu
    )
    x = stone_backsub_lower(y, upper, u)
    return x[:dim]  # Need to trim in case we padded to power of 2.


def stone_triang_lower(
    lower: jnp.ndarray,
    diag: jnp.ndarray,
    upper: jnp.ndarray,
    solve: jnp.ndarray,
    stabilize: bool = True,
    optimized_lu: bool = True,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """
    LU decomposition. `u` is the diagonal of U. `m` is the lower diagonal of `l`.
    The upper diagonal of U is `upper`. The main diagonal of L is 1. Notation follows
    Stone (1973).
    """
    if optimized_lu:
        u, m = _lu(lower, diag, upper, solve, stabilize=stabilize)
    else:
        u, m = _lu_matmul(lower, diag, upper, solve, stabilize=stabilize)
    y = _solve_l(solve, m)
    return u, upper, y


def stone_triang_upper(
    lower: jnp.ndarray,
    diag: jnp.ndarray,
    upper: jnp.ndarray,
    solve: jnp.ndarray,
    stabilize: bool = True,
    optimized_lu: bool = True,
):
    """Triangulate system by removing the upper diagonal."""
    diag, lower, y = stone_triang_lower(
        lower=jnp.flip(upper),
        diag=jnp.flip(diag),
        upper=jnp.flip(lower),
        solve=jnp.flip(solve),
        stabilize=stabilize,
        optimized_lu=optimized_lu,
    )
    return jnp.flip(diag), jnp.flip(lower), jnp.flip(y)


def stone_backsub_upper(
    solve: jnp.ndarray, upper: jnp.ndarray, diag: jnp.ndarray
) -> jnp.ndarray:
    return _solve_u(solve, upper, diag)


def stone_backsub_lower(solve: jnp.ndarray, lower: jnp.ndarray, diag: jnp.ndarray):
    """Backsubstitute to remove the lower diagonal."""
    solution = stone_backsub_upper(jnp.flip(solve), jnp.flip(lower), jnp.flip(diag))
    return jnp.flip(solution)


def _lu_serial(
    lower: jnp.ndarray,
    diag: jnp.ndarray,
    upper: jnp.ndarray,
    solve: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    N = len(solve)
    ui = jnp.zeros_like(diag)
    ui = ui.at[0].set(diag[0])

    for i in range(1, N):
        ui = ui.at[i].set(diag[i] - lower[i - 1] * upper[i - 1] / ui[i - 1])

    mi = jnp.zeros_like(lower)
    mi = mi.at[0].set(lower[0] / diag[0])

    for i in range(1, N - 1):
        mi = mi.at[i].set(lower[i] / (diag[i] - upper[i - 1] * mi[i - 1]))

    return ui, mi


def _lu(
    lower: jnp.ndarray,
    diag: jnp.ndarray,
    upper: jnp.ndarray,
    solve: jnp.ndarray,
    stabilize: bool = True,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    N = len(solve)
    E = lower
    F = upper
    D = diag
    EF = jnp.concatenate([jnp.asarray([0.0]), -E * F])
    QIM2 = jnp.ones((N + 2))
    QIM1 = jnp.concatenate([jnp.asarray([1.0]), diag])
    QI = diag
    QI = QI.at[1:].set(diag[1:] * diag[:-1] + EF[1:])
    temp = jnp.zeros_like(diag)
    U = jnp.zeros_like(diag)
    M = jnp.zeros_like(upper)

    i = 2
    while i <= N / 2:
        if stabilize:
            av_qi = jnp.mean(QI)
            QIM2 /= av_qi
            QIM1 /= av_qi
            QI /= av_qi

        temp = temp.at[i - 2 : N].set(
            QIM1[i - 1 : N + 1] * QIM1[: N - i + 2]
            + EF[: N - i + 2] * QIM2[i : N + 2] * QIM2[: N - i + 2]
        )
        QIM1 = QIM1.at[i : N + 1].set(
            QI[i - 1 : N] * QIM1[: N - i + 1]
            + EF[: N - i + 1] * QIM1[i : N + 1] * QIM2[: N - i + 1]
        )
        QIM2 = QIM2.at[i : N + 2].set(temp[i - 2 : N])
        QI = QI.at[i:N].set(D[i:N] * QIM1[i:N] + EF[i:N] * QIM2[i:N])

        i += i

    U = U.at[:N].set(QI[:N] / QIM1[:N])
    M = M.at[: N - 1].set(E[: N - 1] / U[: N - 1])

    return U, M


def _lu_matmul(
    lower: jnp.ndarray,
    diag: jnp.ndarray,
    upper: jnp.ndarray,
    solve: jnp.ndarray,
    stabilize: bool = True,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Implements an alternative way for computing the LU decomposition."""
    N = len(solve)
    E = lower
    M = jnp.zeros_like(upper)

    a_vals = jnp.zeros((N - 1, 2, 2))
    a_vals = a_vals.at[:, 0, 0].set(diag[1:])
    a_vals = a_vals.at[:, 0, 1].set(-upper * lower)
    a_vals = a_vals.at[:, 1, 0].set(1.0)

    i = 1
    while i < N:
        product = jnp.einsum("bij, bjk -> bik", a_vals[i : N - 1], a_vals[: N - i - 1])
        a_vals = a_vals.at[i : N - 1].set(product)

        if stabilize:
            a_vals = jnp.einsum(
                "bij, b -> bij", a_vals, 1 / jnp.mean(a_vals, axis=(1, 2))
            )

        i += i

    q_init = jnp.asarray([1.0, diag[0]])
    qs = jnp.einsum("bij, j -> bi", a_vals, q_init)

    qs = jnp.concatenate([jnp.asarray([[diag[0], 1.0]]), qs])
    U = qs[:, 0] / qs[:, 1]

    M = M.at[: N - 1].set(E[: N - 1] / U[: N - 1])
    return U, M


def _solve_l(solve: jnp.ndarray, m: jnp.ndarray) -> jnp.ndarray:
    N = len(solve)
    Yi = solve
    Mi = -m
    i = 1

    while i <= N / 2:
        Yi = Yi.at[i:N].set(Yi[i:N] + Yi[: N - i] * Mi[i - 1 : N - 1])
        Mi = Mi.at[i : N - 1].set(Mi[i : N - 1] * Mi[: N - i - 1])
        i += i

    return Yi


def _solve_u(solve: jnp.ndarray, upper: jnp.ndarray, diag: jnp.ndarray) -> jnp.ndarray:
    N = len(diag)

    upper_diag_1 = jnp.asarray(upper / diag[:-1])
    yi_diag_1 = solve / diag

    Yi_2 = yi_diag_1
    Mi_2 = -upper_diag_1

    i = 1

    while i <= N / 2:
        Yi_2 = Yi_2.at[: N - i].set(Yi_2[: N - i] + Yi_2[i:N] * Mi_2[: N - i])
        Mi_2 = Mi_2.at[: N - i - 1].set(Mi_2[: N - i - 1] * Mi_2[i : N - 1])

        i += i

    return Yi_2
